# -*- coding: utf-8 -*-
"""
    crimson.utils
    ~~~~~~~~~~~~~

    General utilities.

    :copyright: (c) 2015 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import json
import re
import sys
from contextlib import contextmanager
from os import linesep

import click
import yaml


if sys.version_info[0] > 2:
    basestring = str


RE_INT = re.compile(r"^([-+]?\d+)L?$")
RE_FLOAT = re.compile(r"^([-+]?\d*\.?\d+(?:[eE][-+]?[0-9]+)?)$")


def convert(raw_str):
    """Tries to convert a string to an int, float, or return it unchanged.

    The function tries first to convert to an int. If it fails, it tries
    to convert to a float. If it fails, the origin input string is returned.

    :param raw_str: Input string.
    :type raw_str: str

    """
    maybe_int = RE_INT.search(raw_str)
    if maybe_int is not None:
        return int(maybe_int.group(1))
    maybe_float = RE_FLOAT.search(raw_str)
    if maybe_float is not None:
        return float(maybe_float.group(1))
    return raw_str


def write_output(payload, out_handle, fmt="json", compact=False, indent=4):
    """Writes the given dictionary as JSON or YAML to the output handle.

    The output handle must have the ``write`` method.

    :param payload: Payload to write.
    :type payload: dict
    :param out_handle: Output handle.
    :type out_handle: object with ``write`` method.
    :param fmt: Output format.
    :type fmt: str (``json`` or ``yaml``)
    :param compact: Whether to write a compact JSON or not. Ignored if the
                    output format is JSON.
    :type compact: bool
    :param indent: Indentation level (ignored if output ``compact`` is true).
    :type indent: int

    """
    if fmt == "json":
        if compact:
            json.dump(payload, out_handle, sort_keys=True, indent=None,
                      separators=(",", ":"))
        else:
            json.dump(payload, out_handle, sort_keys=True, indent=indent)
            out_handle.write(linesep)
    else:
        out_handle.write(yaml.dump(payload, default_flow_style=False,
                         indent=indent))


@contextmanager
def get_handle(input, encoding=None, mode="r"):
    """Context manager for opening files.

    This function returns a file handle of the given file name. You may also
    give an open file handle, in which case the file handle will be yielded
    immediately. The purpose of this is to allow the context manager handle
    both file objects and file names as inputs.

    If a file handle is given, it is not closed upon exiting the context.
    If a file name is given, it will be closed upon exit.

    :param input: Handle of open file or file name.
    :type input: file handle or obj
    :param encoding: Encoding of the file. Ignored if input is file handle.
    :type encoding: str
    :param mode: Mode for opening file. Ignored if input is file handle.
    :type mode: str

    """
    if isinstance(input, basestring):
        assert isinstance(input, basestring), \
            "Unexpected input type: " + repr(input)
        fh = click.open_file(input, mode=mode, encoding=encoding)
    else:
        fh = input

    yield fh

    if isinstance(input, basestring):
        fh.close()



# -*- coding: utf-8 -*-
"""
    crimson.vep
    ~~~~~~~~~~~

    VEP plain text statistics file output parsing.

    :copyright: (c) 2016 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import click

from .utils import convert, get_handle


__all__ = ["parse"]


# Default maximum file size ~ because our parser is not a streaming parser.
_MAX_SIZE = 1024 * 500  # 500 Kb


def group2entry(group):
    """Given the raw string of a VEP statistics group, parse it into a
    key, value tuple.

    Basically, turning this:

        [Variant classes]
        deletion    18
        insertion   35
        SNV 448

    into this:

        ("Variant classes", {
            "deletion": 18,
            "insertion": 35,
            "SNV": 448
         })

    except for then the VEP group represents a histogram, which is the
    case for groups whose name starts with "Distribution of variants ...".

    Those groups, which may look like this:

        [Distribution of variants on chromosome 1]
        0   1
        1   40
        2   35
        3   50
        ...

    gets turned into this:

        ("Distribution of variants on chromosome 1",
         [1, 40, 35, 50, ...])

    """
    raw_key, raw_value = group.split("\n", 1)
    key = raw_key[1:-1]

    values = (line.split("\t", 1) for line in raw_value.strip().split("\n"))

    if not key.startswith("Distribution of variants on"):
        valued = {k: convert(v) for k, v in values}
        return key, valued

    return key, [convert(v) for _, v in values]


def parse(in_data, max_size=_MAX_SIZE):
    """Parses a VEP plain text statistics file into a dictionary.

    :param in_data: Input VEP statistics contents.
    :type in_data: str or file handle
    :param max_size: Maximum expected size of input contents.
    :type max_size: int
    :returns: Parsed VEP statistics values.
    :rtype: dict

    """
    with get_handle(in_data) as fh:
        contents = fh.read(max_size)

    if not contents.startswith("[VEP run statistics]"):
        msg = "Unexpected file structure. No contents parsed."
        raise click.BadParameter(msg)

    entries = [group2entry(g) for g in contents.split("\n\n")]
    return dict(entries)



# -*- coding: utf-8 -*-
"""
    crimson
    ~~~~~~~

    :copyright: (c) 2015 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""

RELEASE = True

__version_info__ = ("0", "2", "0")
__version__ = ".".join(__version_info__)
__version__ += "-dev" if not RELEASE else ""

__author__ = "Wibowo Arindrarto"
__contact__ = "bow@bow.web.id"
__homepage__ = "http://bow.web.id"



# -*- coding: utf-8 -*-
"""
    crimson.flagstat
    ~~~~~~~~~~~~~~~~

    Samtools flagstat output parsing.

    :copyright: (c) 2015 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import re
from functools import partial

import click

from .utils import get_handle


__all__ = ["parse"]


_MAX_SIZE = 1024 * 10
_RE_TOTAL = re.compile(r"(\d+) \+ (\d+) in total")
_RE_DUPLICATES = re.compile(r"(\d+) \+ (\d+) duplicates")
_RE_SECONDARY = re.compile(r"(\d+) \+ (\d+) secondary")
_RE_SUPPLIMENTARY = re.compile(r"(\d+) \+ (\d+) supplimentary")
_RE_MAPPED = re.compile(r"(\d+) \+ (\d+) mapped ")
_RE_PAIRED_SEQ = re.compile(r"(\d+) \+ (\d+) paired in ")
_RE_READ1 = re.compile(r"(\d+) \+ (\d+) read1")
_RE_READ2 = re.compile(r"(\d+) \+ (\d+) read2")
_RE_PAIRED_PROPER = re.compile(r"(\d+) \+ (\d+) properly")
_RE_PAIRED_BAM = re.compile(r"(\d+) \+ (\d+) with itself and")
_RE_SINGLETON = re.compile(r"(\d+) \+ (\d+) singletons")
_RE_DIFF = re.compile(r"(\d+) \+ (\d+) with mate mapped "
                      "to a different chr\s\d")
_RE_DIFF_MIN = re.compile(r"(\d+) \+ (\d+) with mate mapped "
                          "to a different chr\s\(")


def search(text, pattern, caster=str):
    """Searches a text for a pattern and returns the results as the given type.

    :param text: Text to search against.
    :type text: str.
    :param pattern: Compiled regular expression.
    :type pattern: pattern object.
    :param caster: One-argument function that is applied to each search result.
    :type caster: function with one argument.
    :returns: A list of search results.

    """
    search_result = pattern.search(text)
    if search_result is not None:
        return [caster(x) for x in search_result.groups()]
    return [None] * pattern.groups


def parse(in_data):
    """Parses a samtools flagstat result into a dictionary.

    :param in_data: Input flagstat contents.
    :type in_data: str or file handle
    :returns: Parsed flagstat values.
    :rtype: dict

    """
    with get_handle(in_data) as fh:
        contents = fh.read(_MAX_SIZE)

    f = partial(search, contents, caster=int)
    parsed = (
        ("total", f(_RE_TOTAL)),
        ("duplicates", f(_RE_DUPLICATES)),
        ("secondary", f(_RE_SECONDARY)),
        ("supplimentary", f(_RE_SUPPLIMENTARY)),
        ("mapped", f(_RE_MAPPED)),
        ("paired_sequencing", f(_RE_PAIRED_SEQ)),
        ("paired", f(_RE_PAIRED_BAM)),
        ("paired_proper", f(_RE_PAIRED_PROPER)),
        ("read1", f(_RE_READ1)),
        ("read2", f(_RE_READ2)),
        ("singleton", f(_RE_SINGLETON)),
        ("diff_chrom", f(_RE_DIFF)),
        ("diff_chrom_mapq", f(_RE_DIFF_MIN)),
    )
    payload = {
        "pass_qc": {k: v[0] for k, v in parsed if v[0] is not None},
        "fail_qc": {k: v[1] for k, v in parsed if v[1] is not None},
    }
    if len(payload["pass_qc"]) == 0 and len(payload["fail_qc"]) == 0:
        raise click.BadParameter("Cannot parse input flagstat file.")

    return payload



# -*- coding: utf-8 -*-
"""
    crimson.fastqc
    ~~~~~~~~~~~~~~

    FastQC output parsing.

    :copyright: (c) 2015 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
from os import path, walk

import click

from .utils import convert, get_handle


__all__ = ["parse"]


_MAX_LINE_SIZE = 1024


class FastQCModule(object):

    """Class representing a FastQC analysis module."""

    def __init__(self, raw_lines, end_mark='>>END_MODULE'):
        """

        :param raw_lines: list of lines in the module
        :type raw_lines: list of str
        :param end_mark: mark of the end of the module
        :type end_mark: str

        """
        self.extra = {}
        self.raw_lines = raw_lines
        self.end_mark = end_mark
        self.status = None
        self.name = None
        self.contents = self._parse()

    @property
    def dict(self):
        """Module data as a dictionary."""
        payload = {
            "contents": self.contents,
            "status": self.status
        }
        if len(self.extra) > 0:
            for ek, ev in self.extra.items():
                payload[ek] = ev
        return payload

    def _parse(self):
        """Common parser for a FastQC module."""

        # Helper function for converting FastQC values that keeps
        # the "Base" column as strings (since it can be a number
        # or a strin denoting a range)
        def fqc_convert(k, v):
            if k == "Base":
                return v
            return convert(v)

        # check that the last line is a proper end mark
        assert self.raw_lines[-1].startswith(self.end_mark)
        # parse name and status from first line
        tokens = self.raw_lines[0].strip().split('\t')
        name = tokens[0][2:]
        self.name = name
        status = tokens[-1]
        self.status = status
        # the rest of the lines except the last one
        data = []
        if self.name != "Sequence Duplication Levels":
            # and column names from second/third line
            columns = self.raw_lines[1][1:].strip().split("\t")
            self._columns = columns
            for line in self.raw_lines[2:-1]:
                cols = line.strip().split("\t")
                data.append(cols)
        else:
            extra_k, extra_v = self.raw_lines[1][1:].strip().split("\t")
            self.extra[extra_k] = convert(extra_v)
            columns = self.raw_lines[2][1:].strip().split("\t")
            self._columns = columns
            for line in self.raw_lines[3:-1]:
                cols = line.strip().split("\t")
                data.append(cols)

        # optional processing for different modules
        if self.name == 'Basic Statistics':
            data = {k: convert(v) for k, v in data}
        else:
            # zip column names and its values ~ each item in array == one row
            data = [zip(columns, [v for v in d]) for d in data]
            # try to convert numbers appropriately
            # except for "Base" column, since FastQC may output it as range
            data = [{k: fqc_convert(k, v) for k, v in zpd} for zpd in data]

        return data


class FastQC(object):

    """Class representing results from a FastQC run."""

    _mod_names = [
        ">>Basic Statistics",
        ">>Per base sequence quality",
        ">>Per sequence quality scores",
        ">>Per base sequence content",
        ">>Per base GC content",
        ">>Per sequence GC content",
        ">>Per base N content",
        ">>Sequence Length Distribution",
        ">>Sequence Duplication Levels",
        ">>Overrepresented sequences",
        ">>Kmer Content",
    ]

    _mod_map = {k: k.lstrip(">") for k in _mod_names}

    def __init__(self, fp):
        """

        :param fp: open file handle pointing to the FastQC data file
        :type fp: file handle

        """
        self.modules = {}

        line = fp.readline(_MAX_LINE_SIZE)
        attr = ""
        while True:

            tokens = line.strip().split('\t')
            # break on EOF
            if not line:
                break
            # parse version
            elif line.startswith('##FastQC'):
                self.version = line.strip().split()[1]
            # parse individual modules
            elif tokens[0] in self._mod_map:
                attr = self._mod_map[tokens[0]]
                raw_lines = self._read_module(fp, line)
                self.modules[attr] = FastQCModule(raw_lines)

            line = fp.readline(_MAX_LINE_SIZE)

    def _read_module(self, fp, line):
        """Returns a list of lines in a module.

        :param fp: open file handle pointing to the FastQC data file
        :type fp: file handle
        :param line: first line in the module
        :type line: str
        :returns: a list of lines in the module
        :rtype: list of str

        """
        raw = [line]
        while not line.startswith('>>END_MODULE'):
            line = fp.readline(_MAX_LINE_SIZE)
            raw.append(line)

            if not line:
                raise ValueError("Unexpected end of file in module %r" % line)

        return raw

    @property
    def dict(self):
        """FastQC data as a dictionary."""
        payload = {k: v.dict for k, v in self.modules.items()}
        payload["version"] = self.version
        return payload


def parse(in_data):
    """Parses FastQC results into a dictionary.

    :param in_data: File handle of a fastqc_data.txt file, or path to a
                    fastqc_data.txt file or path to a FastQC results directory.
    :type in_data: str or file handle
    :returns: Parsed FastQC values.
    :rtype: dict

    """
    if path.isdir(in_data):
        try:
            ori = in_data
            in_data = path.join(ori, next(walk(ori))[1][0], "fastqc_data.txt")
        except IndexError:
            raise click.BadParameter("Cannot find fastqc_data.txt file in "
                                     "the given directory.")
    with get_handle(in_data) as fh:
        return FastQC(fh).dict



# -*- coding: utf-8 -*-
"""
    crimson.star
    ~~~~~~~~~~~~

    STAR Log.final.out file.

    :copyright: (c) 2016 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import os

import click

from .utils import convert, get_handle


__all__ = ["parse"]


def _pct_convert(raw_str):
    """Converts the given number ending with '%' to a number."""
    if raw_str.endswith("%"):
        return convert(raw_str[:-1])
    return convert(raw_str)


_MAX_SIZE = 1024 * 10
# Mapping of STAR attribute names to python dictionary attribute names
# and the function used to parse it.
_PARSE_MAP = {
    "Started job on":
        ("timeJobStart", str),
    "Started mapping on":
        ("timeMappingStart", str),
    "Finished on":
        ("timeEnd", str),
    "Mapping speed, Million of reads per hour":
        ("mappingSpeed", convert),
    "Number of input reads":
        ("nInput", convert),
    "Average input read length":
        ("avgInputLength", convert),
    "Uniquely mapped reads number":
        ("nUniquelyMapped", convert),
    "Uniquely mapped reads %":
        ("pctUniquelyMapped", _pct_convert),
    "Average mapped length":
        ("avgMappedLength", convert),
    "Number of splices: Total":
        ("nSplicesTotal", convert),
    "Number of splices: Annotated (sjdb)":
        ("nSplicesAnnotated", convert),
    "Number of splices: GT/AG":
        ("nSplicesGTAG", convert),
    "Number of splices: GC/AG":
        ("nSplicesGCAG", convert),
    "Number of splices: AT/AC":
        ("nSplicesATAC", convert),
    "Number of splices: Non-canonical":
        ("nSplicesNonCanonical", convert),
    "Mismatch rate per base, %":
        ("rateMismatchPerBase", _pct_convert),
    "Deletion rate per base":
        ("rateDeletionPerBase", _pct_convert),
    "Deletion average length":
        ("avgDeletionLength", convert),
    "Insertion rate per base":
        ("rateInsertionPerBase", _pct_convert),
    "Insertion average length":
        ("avgInsertionLength", convert),
    "Number of reads mapped to multiple loci":
        ("nMappedMultipleLoci", convert),
    "% of reads mapped to multiple loci":
        ("pctMappedMultipleLoci", _pct_convert),
    "Number of reads mapped to too many loci":
        ("nMappedTooManyLoci", convert),
    "% of reads mapped to too many loci":
        ("pctMappedTooManyLoci", _pct_convert),
    "% of reads unmapped: too many mismatches":
        ("pctUnmappedForTooManyMismatches", _pct_convert),
    "% of reads unmapped: too short":
        ("pctUnmappedForTooShort", _pct_convert),
    "% of reads unmapped: other":
        ("pctUnmappedForOther", _pct_convert),
}


def parse(in_data):
    """Parses the log of a STAR run.

    :param in_data: Input STAR-Fusion contents.
    :type in_data: str or file handle
    :returns: Parsed values.
    :rtype: dict

    """
    payload = {}
    with get_handle(in_data) as src:
        contents = src.read(_MAX_SIZE)

    for line in contents.split(os.linesep):
        # pass empty lines
        if line and not line.strip():
            continue
        line = line.strip()
        if '|' in line:
            ori_key, val = [x.strip() for x in line.split('|', 1)]
            if ori_key in _PARSE_MAP:
                key, func = _PARSE_MAP[ori_key]
                val = func(val)
                if key in payload:
                    msg = "Unexpected duplicate key entry: {0} ({1})."
                    raise click.BadParameter(msg.format(key, ori_key))
                payload[key] = val

    if not payload:
        msg = "Unexpected file structure. No contents parsed."
        raise click.BadParameter(msg)
    return payload



# -*- coding: utf-8 -*-
"""
    crimson.main
    ~~~~~~~~~~~~

    Main entry point for command line invocation.

    :copyright: (c) 2015 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import click

from . import __version__
from . import fastqc as m_fastqc
from . import flagstat as m_flagstat
from . import picard as m_picard
from . import star as m_star
from . import star_fusion as m_star_fusion
from . import vep as m_vep
from .utils import write_output

__all__ = []


@click.group()
@click.version_option(__version__)
@click.option("--fmt", default="json", type=click.Choice(["json", "yaml"]),
              help="Output file format. Default: json.")
@click.option("--indent", default=2,
              help="Indentation level. Ignored if the --compact flag is set. "
              "Default: 2.")
@click.option("--compact", is_flag=True,
              help="Whether to create a compact JSON or not. "
              "Ignored if output format is YAML.")
@click.pass_context
def cli(ctx, fmt, indent, compact):
    """Converts bioinformatics tools' output to a standard format."""
    ctx.params["fmt"] = fmt
    ctx.params["indent"] = indent
    ctx.params["compact"] = compact


@cli.command()
@click.argument("input", type=click.Path(exists=True))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def fastqc(ctx, input, output):
    """Converts FastQC output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_fastqc.parse(input)
    write_output(payload, output, **ctx.parent.params)


@cli.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def flagstat(ctx, input, output):
    """Converts samtools flagstat output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_flagstat.parse(input)
    write_output(payload, output, **ctx.parent.params)


@cli.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def picard(ctx, input, output):
    """Converts Picard metrics output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_picard.parse(input)
    write_output(payload, output, **ctx.parent.params)


@cli.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def star(ctx, input, output):
    """Converts STAR Log.final.out file.

    Use "-" for stdin and/or stdout.

    """
    payload = m_star.parse(input)
    write_output(payload, output, **ctx.parent.params)


@cli.command(name="star-fusion")
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def star_fusion(ctx, input, output):
    """Converts output of STAR-Fusion.

    Use "-" for stdin and/or stdout.

    """
    payload = m_star_fusion.parse(input)
    write_output(payload, output, **ctx.parent.params)


@cli.command(name="vep")
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def vep(ctx, input, output):
    """Converts plain text output of Variant Effect Predictor.

    Use "-" for stdin and/or stdout.

    """
    payload = m_vep.parse(input)
    write_output(payload, output, **ctx.parent.params)



# -*- coding: utf-8 -*-
"""
    crimson.star_fusion
    ~~~~~~~~~~~~~~~~~~~

    STAR-Fusion output parsing.

    :copyright: (c) 2016 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import click

from .utils import get_handle


__all__ = ["parse"]

# Expected column names
# Abridged column names
_ABR_COLS = [
    "fusion_name", "JunctionReads", "SpanningFrags", "Splice_type",
    "LeftGene", "LeftBreakpoint",
    "RightGene", "RightBreakpoint",
]
# Non-abridged column names
_NONABR_COLS = [
    "fusion_name", "JunctionReads", "SpanningFrags", "Splice_type",
    "LeftGene", "LeftBreakpoint",
    "RightGene", "RightBreakpoint",
    "JunctionReads", "SpanningFrags",
]

# Delimiter strings
_DELIM = {
    # Gene name-id
    "gids": "^",
    # Chromosome-coordinate-strand
    "loc": ":",
}


def parse_lr_entry(lr_gene, lr_brkpoint):
    """Parses the gene and breakpoint entry.

    :param lr_gene: Column value for 'LeftGene' or 'RightGene'.
    :type lr_gene: str
    :param lr_brkpoint: Column value for 'LeftBreakpoint' or 'RightBreakpoint'.
    :type lr_brkpoint: str
    :rtype: dict

    """
    lrgname, lrgid = lr_gene.split(_DELIM["gids"])
    lrchrom, lrpos, lrstrand = lr_brkpoint.split(_DELIM["loc"])

    return {
        "geneName": lrgname,
        "geneID": lrgid,
        "chromosome": lrchrom,
        "position": int(lrpos),
        "strand": lrstrand,
    }


def parse_raw_line(raw_line, colnames, is_abridged=True):
    """Parses a single line into a dictionary.

    :param raw_line: STAR-Fusion result line.
    :type raw_line: str
    :param colnames: Column names present in the file.
    :type colnames: list of str
    :param is_abridged: Whether the input raw line is from an abridged file.
    :type is_abridged: bool
    :rtype: dict

    """
    values = raw_line.split("\t")
    if len(values) != len(colnames):
        msg = "Line values {0} does not match column names {1}."
        raise click.BadParameter(msg.format(values, colnames))

    entries, reads = {}, {}
    if is_abridged:
        entries = {k: v for k, v in zip(colnames, values)}
    else:
        entry_colnames, read_colnames = colnames[:-2], colnames[-2:]
        entry_values, read_values = values[:-2], values[-2:]
        entries = {k: v for k, v in zip(entry_colnames, entry_values)}
        reads = {k: v.split(",") for k, v in zip(read_colnames, read_values)}

    ret = {
        "fusionName": entries["fusion_name"],
        "nJunctionReads": int(entries["JunctionReads"]),
        "nSpanningFrags": int(entries["SpanningFrags"]),
        "spliceType": entries["Splice_type"],
        "left": parse_lr_entry(entries["LeftGene"],
                               entries["LeftBreakpoint"]),
        "right": parse_lr_entry(entries["RightGene"],
                                entries["RightBreakpoint"]),
    }
    if reads:
        ret["reads"] = {
            "junctionReads": reads["JunctionReads"],
            "spanningFrags": reads["SpanningFrags"],
        }

    return ret


def parse(in_data):
    """Parses the abridged output of a STAR-Fusion run.

    :param in_data: Input STAR-Fusion contents.
    :type in_data: str or file handle
    :returns: Parsed values.
    :rtype: dict

    """
    payload = []
    with get_handle(in_data) as src:
        first_line = src.readline().strip()
        if not first_line.startswith("#"):
            msg = "Unexpected header line: '{0}'."
            raise click.BadParameter(msg.format(first_line))
        # Parse column names, after removing the '#' character
        colnames = first_line[1:].split("\t")
        if colnames != _ABR_COLS and colnames != _NONABR_COLS:
            msg = "Unexpected column names: {0}."
            raise click.BadParameter(msg.format(colnames))
        else:
            is_abridged = colnames == _ABR_COLS
            for line in (x.strip() for x in src):
                parsed = parse_raw_line(line, colnames, is_abridged)
                payload.append(parsed)

    return payload



# -*- coding: utf-8 -*-
"""
    crimson.picard
    ~~~~~~~~~~~~~~

    Picard metrics file parsing.

    :copyright: (c) 2015 Wibowo Arindrarto <bow@bow.web.id>
    :license: BSD

"""
import os
import re

import click

from .utils import convert, get_handle


_RE_HEADER = re.compile(r"^#+\s+")


__all__ = ["parse"]


def fetch(l, pred, first=True):
    """Fetches item(s) that return true in the given list.

    :param l: Input list to fetch from.
    :type l: list
    :param pred: Predicate function.
    :type pred: function
    :param first: Whether to return only the first matching item or
                  all matching items.
    :type first: bool

    """
    matches = [x for x in l if pred(x)]
    if len(matches) == 0:
        return None
    if first:
        return matches[0]
    return matches


def parse_header(header):
    """Parses the Picard header lines into a dictionary.

    :param header: Raw Picard header string.
    :type header: str
    :returns: Parsed header information.
    :rtype: dict

    """
    parsed = [_RE_HEADER.sub("", x) for x in header.split(os.linesep)]
    if len(parsed) != 4:
        raise ValueError("Unexpected Picard header.")

    return {"flags": parsed[1], "time": parsed[3]}


def parse_metrics(metrics):
    """Parses the Picard metrics lines into a dictionary.

    :param metris: Raw Picard metrics string.
    :type metrics: str
    :returns: Parsed metrics table.
    :rtype: dict

    """
    if metrics is None:
        return

    lines = [l.strip(os.linesep) for l in metrics.split(os.linesep)]

    try:
        metrics_class = lines.pop(0).split("\t")[1]
    except IndexError:
        metrics_class = None

    parsed = []
    for line in lines:
        parsed.append([convert(v) for v in line.split("\t")])

    header_cols = parsed.pop(0)
    contents = [dict(zip(header_cols, l)) for l in parsed]
    if len(contents) == 1:
        contents = contents.pop()
    payload = {"contents": contents}
    if metrics_class is not None:
        payload["class"] = metrics_class

    return payload


def parse_histogram(histo):
    """Parses the Picard histogram lines into a dictionary.

    :param metris: Raw Picard histogram string.
    :type metrics: str
    :returns: Parsed histogram table.
    :rtype: dict

    """
    if histo is None:
        return

    lines = [l.strip(os.linesep) for l in histo.split(os.linesep)]
    lines.pop(0)

    parsed = []
    for line in lines:
        parsed.append([convert(v) for v in line.split("\t")])

    header_cols = parsed.pop(0)
    payload = {"contents": [dict(zip(header_cols, l)) for l in parsed]}

    return payload


def parse(in_data):
    """Parses an input Picard metrics file into a dictionary.

    :param in_data: Input metrics file.
    :type in_data: str or file handle
    :returns: Parsed metrics values.
    :rtype: dict

    """
    with get_handle(in_data) as fh:
        contents = fh.read(1024 * 1024 * 1)

    sections = contents.strip(os.linesep).split(os.linesep * 2)

    header = fetch(sections, lambda x: x.startswith("## htsjdk"))
    if header is None:
        raise click.BadParameter("Unexpected Picard metrics file format.")
    metrics = fetch(sections, lambda x: x.startswith("## METRICS"))
    histo = fetch(sections, lambda x: x.startswith("## HISTOGRAM"))

    return {
        "header": parse_header(header),
        "metrics": parse_metrics(metrics),
        "histogram": parse_histogram(histo),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

from crimson import __author__, __contact__, \
        __homepage__, __version__


with open("README.rst") as src:
    readme = src.read()

with open("HISTORY.rst") as src:
    history = src.read().replace(".. :changelog:", "").strip()

with open("requirements.txt") as src:
    requirements = [line.strip() for line in src]

with open("requirements-dev.txt") as src:
    test_requirements = [line.strip() for line in src]


setup(
    name="Crimson",
    version=__version__,
    description="Bioinformatics tool outputs converter to JSON or YAML.",
    long_description=readme + "\n\n" + history,
    author=__author__,
    author_email=__contact__,
    url=__homepage__,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    entry_points="""
    [console_scripts]
    crimson=crimson.main:cli
    """,
    keywords="crimson bioinformatics json yaml samtools picard fastqc",
    tests_require=test_requirements,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Utilities",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Libraries",
    ],
)
